from utils import get_logger, read_json, get_line_number
import json
from typing import Dict, List, Union, Tuple, NewType

# Anton imports
import constants
import shared_variables
import os

logger = get_logger("AudioFiles")


class TYPES:
    uploaded_file_name = NewType("uploaded_file_name", str)  # this should be the name of the uploaded file

    audio_file_button_id = NewType("audio_file_button_id", str)
    audio_file_button_name = NewType("audio_file_button_name", str)
    audio_file_timestamp = NewType("audio_file_timestamp", str)

    audio_file = Dict[str, Union[uploaded_file_name, audio_file_button_name, audio_file_timestamp]]
    audio_files = Dict[audio_file_button_id, audio_file]


class AudioFiles:
    DIFFERENT_AUDIO_FILES_NUM = set([str(i) for i in range(1, 10)])

    def __init__(self):
        self._audio_files = read_json("audio_files_backup.json")
        self.occupied_audio_files_id = set(self._audio_files.keys())

    @property
    def value(self) -> TYPES.audio_files:
        return self._audio_files

    @property
    def new_id(self) -> Union[None, TYPES.audio_file_button_id]:
        """Gets an unoccupied audio_file id.

        Returns:
            Union[str, None]: returns id or None depending upon availability.
        """
        available_ids = AudioFiles.DIFFERENT_AUDIO_FILES_NUM - self.occupied_audio_files_id
        if len(available_ids) != 0:
            return min(available_ids)
        else:
            return None

    @property
    def bytes(self) -> bytes:
        return json.dumps(self._audio_files).encode()

    def get_audio_files(self, audio_files_ids: List[TYPES.audio_file_button_id]) -> List[TYPES.uploaded_file_name]:
        files_ids = set([])
        for id in audio_files_ids:
            files_ids.update(self._audio_files[str(id)]["file_name"])
        return list(files_ids)

    def validate_audio_files(self, uploaded_files: TYPES.uploaded_file_name) -> Tuple[bool, str]:
        # type check:
        if not isinstance(uploaded_files, str):
            return False, "Not a file name str"
        elif len(uploaded_files) == 0:
            return False, "Empty file name"
        else:
            return True, "OK"

    def add(self, audio_file: TYPES.audio_file) -> Tuple[bool, str]:
        audio_file_button_name = audio_file["btn_name"]  # That would be the Button name
        uploaded_file_name = audio_file["file_name"]
        valid_audio_files, msg = self.validate_audio_files(uploaded_file_name)
        new_id = self.new_id
        logger.warning("Giving new_id: %s", new_id)

        if new_id is not None and valid_audio_files:
            timestamp = shared_variables.audio_processor.addNewFile(
                os.path.join(constants.FILE_PATH, uploaded_file_name))
            logger.info("Finished adding new file")

            self._audio_files[new_id] = {
                "btn_name": audio_file_button_name, "file_name": uploaded_file_name, "timestamp": timestamp
            }
            self.occupied_audio_files_id.add(new_id)
            self.save()
            logger.info(f"File info: {self._audio_files[new_id]}")
            return True, {"id": new_id}
        else:
            logger.warning("Failed adding file!!!")
            return False, "Improper Request"

    def update(self, audio_file: TYPES.audio_file) -> Tuple[bool, str]:
        audio_file_button_name = audio_file["btn_name"]
        uploaded_file_name = audio_file["file_name"]
        audio_file_button_id = str(audio_file["btn_id"])
        # Get stored timestamp unless the file has changed
        timestamp = self._audio_files[audio_file_button_id]["timestamp"]
        logger.info(f"Timestamp from the audio file: {timestamp}")

        valid_audio_files, msg = self.validate_audio_files(uploaded_file_name)
        if audio_file_button_id in self._audio_files:
            if valid_audio_files:
                # Check if the file name has changed. If so, encode the new file and delete the old one
                if uploaded_file_name != self._audio_files[audio_file_button_id]["file_name"]:
                    logger.warning("Updating associated audio file for \"%s\"",
                                   self._audio_files[audio_file_button_id]["btn_name"])
                    logger.info("Removing the old file: %s",
                                self._audio_files[audio_file_button_id]["file_name"])
                    shared_variables.audio_processor. \
                        removeFile(self._audio_files[str(audio_file["btn_id"])][str("file_name")], constants.YES)
                    logger.info("New filename: %s", uploaded_file_name)
                    # Use the returned timestamp to update the json
                    timestamp = shared_variables.audio_processor.addNewFile(
                        os.path.join(constants.FILE_PATH, uploaded_file_name))

                self._audio_files[audio_file_button_id] = {
                    "btn_name": audio_file_button_name, "file_name": uploaded_file_name, "timestamp": timestamp
                }
                self.save()
                logger.info("Finished updating file")
                return True, "OK"
            else:
                return False, "Audio files are not valid"
        else:
            return -1, "Audio file doesn't exist!"

    def remove(self, audio_file: Dict[str, TYPES.audio_file_button_id]) -> Tuple[bool, str]:

        if str(audio_file["btn_id"]) in self._audio_files:

            try:
                shared_variables.audio_processor.\
                    removeFile(self._audio_files[str(audio_file["btn_id"])][str("file_name")], constants.YES)

                del self._audio_files[str(audio_file["btn_id"])]
                logger.warning(f"Remove success: {audio_file['btn_id']}")
                self.refresh_occupied_ids()
                self.save()
                return True, "OK"

            except Exception as e:
                logger.error(f"Error: {e}")
                return False, f"Unable to delete file ID {audio_file['btn_id']}"
        else:
            logger.warning(f"Audio file deletion unsuccessfull, {audio_file['btn_id']} doesn't exist")
            return False, "Audio file doesn't exist"

    def save(self):
        with open("audio_files_backup.json", "w") as f:
            json.dump(self._audio_files, f)

    def refresh_occupied_ids(self):
        self.occupied_audio_files_id = set(self._audio_files.keys())

    def get_audio_file_timestamp(self, audio_files_id):
        return self._audio_files[str(audio_files_id)]["timestamp"]

    def get_timestamp_from_name(self, file_path):
        filename = os.path.splitext(os.path.basename(file_path))[0]+".mp3"
        json_object = read_json("audio_files_backup.json")
        file_dict = {}

        for key in json_object:
            if json_object[key]["file_name"] == filename:
                file_dict["file_id"] = key
                file_dict["timestamp"] = json_object[key]["timestamp"]
                break

        return file_dict

    def update_timestamp(self, file_id, timestamp):
        fid = file_id
        new_timestamp = timestamp
        json_object = read_json("audio_files_backup.json")

        json_object[str(fid)]["timestamp"] = new_timestamp
        a_file = open("audio_files_backup.json", "w")
        json.dump(json_object, a_file)
        a_file.close()
        logger.info("File updated.")



