from BARP.barp import BARP
import time
import socket
from utils import is_multicast, get_logger

# import alsaaudio
from datetime import datetime
import struct
import logging
import shared_variables

# Anton imports
import constants

logger = get_logger("PagingMaster")


class paging_master:
    """
    Class defining basic IC paging master attributes.
    """

    def __init__(self, id):
        self.barp = BARP(id)
        # config = read_json("config.json")
        # Use shared config.
        config = shared_variables.config
        self.header_kwargs = config["barp_header_kwargs"]
        self.header_kwargs["sender_id"] = id
        self._ports = config["ports"]
        self._audio_device = config["audio_device"]
        self._encoding_type = config["encoding_type"]
        self._multicast_group_address = config["multicast_group_address"]
        self._state = 0  # starts with IDLE state.
        self._barp_ip = config["barp_ip"]
        self.active_devices = {}
        self.listen_status_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.status_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        self.control_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        if is_multicast(self.barp_ip):
            self.control_socket.setsockopt(
                socket.SOL_SOCKET, socket.IP_MULTICAST_TTL, 1
            )
            self.status_socket.setsockopt(socket.SOL_SOCKET, socket.IP_MULTICAST_TTL, 1)
        else:
            self.control_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            self.status_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.audio_socket = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        if is_multicast(self._multicast_group_address):
            self.audio_socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
        else:
            logging.warning(
                "{} is not multicast,audio will be bradcasted instead".format(
                    self._multicast_group_address
                )
            )
            self._multicast_group_address = "192.168.10.255"
            self.audio_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._exit = False
        # Constants
        self.PAGING_REQUEST_TIME_INTERVAL = 3
        self.STATUS_REPORT_TIME_INTERVAL = 5

    @property
    def exit_flag(self):
        return self._exit

    @property
    def state(self):
        """
        Possibilites:
        0(IDLE) and 130(Sending the message)
        """
        return self._state

    @property
    def control_port(self):
        """
        Port used for sending the control messages.
        """
        return self._ports["control"]

    @property
    def audio_port(self):
        """
        Port used for sending the audio message/stream/call
        """
        return self._ports["audio"]

    @property
    def status_port(self):
        """
        Port used for sending the  status message
        """
        return self._ports["status"]

    @property
    def encoding_type(self):
        """
        Encoding type to be used for the audio message.
        """
        return self._encoding_type

    @property
    def multicast_group(self):
        """
        Mutlticast group address. To be used while sending the paging message.
        """
        if is_multicast(self._multicast_group_address):
            return self._multicast_group_address
        else:
            return "<broadcast>"

    @property
    def record_device(self):
        return self._audio_device["record"]

    @property
    def playback_device(self):
        return self._audio_device["playback"]

    @property
    def barp_ip(self):
        return self._barp_ip

    def exit(self):
        self._exit = True

    def listen_status_message(self):
        """
        To be started as a parallel thread process for
        listening to active devices.
        """
        if is_multicast(self.barp_ip):

            self.listen_status_socket.bind(("", self.status_port))
            mreq = struct.pack("4sL", self.barp_ip, socket.INADDR_ANY)
            self.listen_status_socket.setsockopt(
                socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq
            )

        else:
            # Switching to listening on broadacast port if ip is not multicast.
            self.listen_status_socket.bind((self.barp_ip, self.status_port))
        self.listen_status_socket.settimeout(1)
        while not self.exit_flag:

            try:
                data, address = self.listen_status_socket.recvfrom(4096)
            except socket.timeout:
                continue
            msg = self.barp.decode(data)
            if msg[0]["message_type"] == 128:
                self.active_devices[msg[0]["sender_id"]] = {
                    "state": msg[1]["state"],
                    "time_stamp": datetime.now(),
                }

            # Removal of devices with stale status.
            current_time = datetime.now()
            for dev_id in list(self.active_devices.keys()):
                if (
                    current_time - self.active_devices[dev_id]["time_stamp"]
                ).seconds > 5:
                    del self.active_devices[dev_id]
        logging.info("Status listening stopped!")

    def send_status_message(self):
        """
        To be started as a parallel thread process.
        Prepares a state message for the paging master.
        In this POC, only state 0 (Idle) and 130 (paging message)
        for other possible states please refer to the barp.pdf doc
        """
        while not self.exit_flag:
            message_kwargs = {
                "state": self.state,
                "priority": 0,
                "sender_id": self.barp.my_id,
                "volume": 50,
                "config_bgm_channel": 0,
            }
            status = self.barp.encode(
                message_type=128, message_kwargs=message_kwargs, **self.header_kwargs
            )
            self.status_socket.sendto(status, (self.barp_ip, self.status_port))
            time.sleep(self.STATUS_REPORT_TIME_INTERVAL)
        logging.info("Status message sending finished")

    def send_page_group_request(self, clients, priority=0):
        """
        Prepares and sends state message 3 to be sent to the client
        before starting the stream.
        Args:
            clients (List[int])
            priority (int)
        """
        self._state = 130
        while self._state == 130:
            message_kwargs = {
                "port": self.audio_port,
                "priority": priority,
                "audio_encdoing_type": self.encoding_type,
                "multicast_stream_group": self.multicast_group
                if is_multicast(self._multicast_group_address)
                else "0.0.0.0",
                "receiver_id": clients,
            }
            request_msg = self.barp.encode(
                message_type=3, message_kwargs=message_kwargs, **self.header_kwargs
            )

            if shared_variables.playfile_state == constants.BUSY:
                # print(f"Playing file to client {clients}")
                logger.info(f"Playing file to client {clients}")
            else:
                # print(f"Paging to client {clients}")
                logger.info(f"Paging to client {clients}")

            if is_multicast(self.barp_ip):
                self.control_socket.sendto(
                    request_msg, (self.barp_ip, self.control_port)
                )
            else:
                self.control_socket.sendto(
                    request_msg, ("<broadcast>", self.control_port)
                )
            time.sleep(self.PAGING_REQUEST_TIME_INTERVAL)

    def send_end_group_request(self, clients):
        """
        Prepares and sends the end group page request, MSG 8.
        The stream should stop after this message is sent.
        Args:
            clients (List[int])
        """
        self._state = 0
        message_kwargs = {"port": self.audio_port, "receiver_id": clients}
        end_page_request = self.barp.encode(
            message_type=8, message_kwargs=message_kwargs, **self.header_kwargs
        )
        if shared_variables.playfile_state == constants.BUSY:
            # print(f"Stopping playing file to client {clients}")
            logger.info(f"Stopping playing file to client {clients}")
        else:
            # print(f"Stopping paging to client {clients}")
            logger.info(f"Stopping paging to client {clients}")

        if is_multicast(self.barp_ip):
            self.control_socket.sendto(
                end_page_request, (self.barp_ip, self.control_port)
            )
        else:
            self.control_socket.sendto(
                end_page_request, ("<broadcast>", self.control_port)
            )
