
// Put this file in /current/javascript directory
'use strict';
class Resampler {
  constructor(fromSampleRate, toSampleRate, channels, inputBufferSize) {

    if (!fromSampleRate || !toSampleRate || !channels) {
      throw (new Error("Invalid settings specified for the resampler."));
    }
    this.resampler = null;
    this.fromSampleRate = fromSampleRate;
    this.toSampleRate = toSampleRate;
    this.channels = channels || 0;
    this.inputBufferSize = inputBufferSize;
    this.initialize()
  }

  initialize() {
    if (this.fromSampleRate == this.toSampleRate) {

      // Setup resampler bypass - Resampler just returns what was passed through
      this.resampler = (buffer) => {
        return buffer
      };
      this.ratioWeight = 1;

    } else {

      if (this.fromSampleRate < this.toSampleRate) {

        // Use generic linear interpolation if upsampling,
        // as linear interpolation produces a gradient that we want
        // and works fine with two input sample points per output in this case.
        this.linearInterpolation();
        this.lastWeight = 1;

      } else {

        // Custom resampler I wrote that doesn't skip samples
        // like standard linear interpolation in high downsampling.
        // This is more accurate than linear interpolation on downsampling.
        this.multiTap();
        this.tailExists = false;
        this.lastWeight = 0;
      }

      // Initialize the internal buffer:
      this.initializeBuffers();
      this.ratioWeight = this.fromSampleRate / this.toSampleRate;
    }
  }

  bufferSlice(sliceAmount) {

    //Typed array and normal array buffer section referencing:
    try {
      return this.outputBuffer.subarray(0, sliceAmount);
    }
    catch (error) {
      try {
        //Regular array pass:
        this.outputBuffer.length = sliceAmount;
        return this.outputBuffer;
      }
      catch (error) {
        //Nightly Firefox 4 used to have the subarray function named as slice:
        return this.outputBuffer.slice(0, sliceAmount);
      }
    }
  }

  initializeBuffers() {
    this.outputBufferSize = (Math.ceil(this.inputBufferSize * this.toSampleRate / this.fromSampleRate / this.channels * 1.000000476837158203125) + this.channels) + this.channels;
    try {
      this.outputBuffer = new Float32Array(this.outputBufferSize);
      this.lastOutput = new Float32Array(this.channels);
    }
    catch (error) {
      this.outputBuffer = [];
      this.lastOutput = [];
    }
  }

  linearInterpolation() {
    this.resampler = (buffer) => {
      let bufferLength = buffer.length,
        channels = this.channels,
        outLength,
        ratioWeight,
        weight,
        firstWeight,
        secondWeight,
        sourceOffset,
        outputOffset,
        outputBuffer,
        channel;

      if ((bufferLength % channels) !== 0) {
        throw (new Error("Buffer was of incorrect sample length."));
      }
      if (bufferLength <= 0) {
        return [];
      }

      outLength = this.outputBufferSize;
      ratioWeight = this.ratioWeight;
      weight = this.lastWeight;
      firstWeight = 0;
      secondWeight = 0;
      sourceOffset = 0;
      outputOffset = 0;
      outputBuffer = this.outputBuffer;

      for (; weight < 1; weight += ratioWeight) {
        secondWeight = weight % 1;
        firstWeight = 1 - secondWeight;
        this.lastWeight = weight % 1;
        for (channel = 0; channel < this.channels; ++channel) {
          outputBuffer[outputOffset++] = (this.lastOutput[channel] * firstWeight) + (buffer[channel] * secondWeight);
        }
      }
      weight -= 1;
      for (bufferLength -= channels, sourceOffset = Math.floor(weight) * channels; outputOffset < outLength && sourceOffset < bufferLength;) {
        secondWeight = weight % 1;
        firstWeight = 1 - secondWeight;
        for (channel = 0; channel < this.channels; ++channel) {
          outputBuffer[outputOffset++] = (buffer[sourceOffset + ((channel > 0) ? (channel) : 0)] * firstWeight) + (buffer[sourceOffset + (channels + channel)] * secondWeight);
        }
        weight += ratioWeight;
        sourceOffset = Math.floor(weight) * channels;
      }
      for (channel = 0; channel < channels; ++channel) {
        this.lastOutput[channel] = buffer[sourceOffset++];
      }
      return this.bufferSlice(outputOffset);
    };
  }

  multiTap() {
    this.resampler = (buffer) => {
      let bufferLength = buffer.length,
        outLength,
        output_variable_list,
        channels = this.channels,
        ratioWeight,
        weight,
        channel,
        actualPosition,
        amountToNext,
        alreadyProcessedTail,
        outputBuffer,
        outputOffset,
        currentPosition;

      if ((bufferLength % channels) !== 0) {
        throw (new Error("Buffer was of incorrect sample length."));
      }
      if (bufferLength <= 0) {
        return [];
      }

      outLength = this.outputBufferSize;
      output_variable_list = [];
      ratioWeight = this.ratioWeight;
      weight = 0;
      actualPosition = 0;
      amountToNext = 0;
      alreadyProcessedTail = !this.tailExists;
      this.tailExists = false;
      outputBuffer = this.outputBuffer;
      outputOffset = 0;
      currentPosition = 0;

      for (channel = 0; channel < channels; ++channel) {
        output_variable_list[channel] = 0;
      }

      do {
        if (alreadyProcessedTail) {
          weight = ratioWeight;
          for (channel = 0; channel < channels; ++channel) {
            output_variable_list[channel] = 0;
          }
        } else {
          weight = this.lastWeight;
          for (channel = 0; channel < channels; ++channel) {
            output_variable_list[channel] = this.lastOutput[channel];
          }
          alreadyProcessedTail = true;
        }
        while (weight > 0 && actualPosition < bufferLength) {
          amountToNext = 1 + actualPosition - currentPosition;
          if (weight >= amountToNext) {
            for (channel = 0; channel < channels; ++channel) {
              output_variable_list[channel] += buffer[actualPosition++] * amountToNext;
            }
            currentPosition = actualPosition;
            weight -= amountToNext;
          } else {
            for (channel = 0; channel < channels; ++channel) {
              output_variable_list[channel] += buffer[actualPosition + ((channel > 0) ? channel : 0)] * weight;
            }
            currentPosition += weight;
            weight = 0;
            break;
          }
        }

        if (weight === 0) {
          for (channel = 0; channel < channels; ++channel) {
            outputBuffer[outputOffset++] = output_variable_list[channel] / ratioWeight;
          }
        } else {
          this.lastWeight = weight;
          for (channel = 0; channel < channels; ++channel) {
            this.lastOutput[channel] = output_variable_list[channel];
          }
          this.tailExists = true;
          break;
        }
      } while (actualPosition < bufferLength && outputOffset < outLength);
      return this.bufferSlice(outputOffset);
    };
  }

  resample(buffer) {
    if (this.fromSampleRate == this.toSampleRate) {
      this.ratioWeight = 1;
    } else {
      if (this.fromSampleRate < this.toSampleRate) {
        this.lastWeight = 1;
      } else {
        this.tailExists = false;
        this.lastWeight = 0;
      }
      this.initializeBuffers();
      this.ratioWeight = this.fromSampleRate / this.toSampleRate;
    }
    return this.resampler(buffer)
  }
};

function convertFloat32ToPcm16(float32Buffer, my_id) {
  // Converts the raw recorded audio from 32bit float to pcm16bit.
  let l = float32Buffer.length + my_id.length;  //Buffer
  let int16Buffer = new Int16Array(l);
  for (let idx = 0; idx < my_id.length; idx++) {
    int16Buffer[idx] = my_id.charCodeAt(idx);
  }
  for (let idx = my_id.length; idx < l; idx++) {
    let s = Math.max(-1, Math.min(1, float32Buffer[idx - my_id.length]));
    int16Buffer[idx] = s < 0 ? s * 0x8000 : s * 0x7FFF;
    // int16Buffer[idx] = float32Buffer[idx - my_id.length]
  }
  return int16Buffer;
};

let my_id = null

const PARAMS = {
  "sampleRate": 24000,  // 24000
  "bufferLength": 16 * 1024, // Value in kb, only power of 2
  "channels": 1,
  "https_audio_Server": ""
}

let page_button_function = "default";

const set_buttons = function (state) {
  if (state == 'inactive') {
    document.getElementById('page').disabled = true
    document.getElementById('page').classList.add("disabled")
    page_button_function = null;
  }
  if (state == 'paging_enabled') {
    //paging active
    page_button_function = 'page'
    document.getElementById('page').classList.remove("disabled")
    document.getElementById('page').disabled = false
    document.getElementById('page').classList.add("pressed")
    document.getElementById('page').textContent = "Release to STOP"
  }
  if (state == 'paging_idle') {
    //paging idle
    page_button_function = 'page'
    document.getElementById('page').classList.remove("disabled")
    document.getElementById('page').disabled = false
    document.getElementById('page').classList.remove("pressed")
    document.getElementById('page').textContent = "Push to TALK"
  }
  if (state == 'recording_idle') {
    //start recording page message
    page_button_function = 'record'
    document.getElementById('page').classList.remove("disabled")
    document.getElementById('page').disabled = false
    document.getElementById('page').classList.remove("pressed")
    document.getElementById('page').textContent = "Push to record"
  }
  if (state == 'recording_start') {
    //stop recording page message
    page_button_function = 'record'
    document.getElementById('page').classList.remove("disabled")
    document.getElementById('page').disabled = false
    document.getElementById('page').classList.add("pressed")
    document.getElementById('page').textContent = "Release to STOP"    
  }
};

//Takes input params as our original function and milliseconds (ms) to wait.
const debounce = (fn, wait) => {
  let timeout;
  // Returns a new function
  return (args) => {
      // Clears any exiting timeout. This ensures every time the function is called again a new timeout is created.
      clearTimeout(timeout);

      // Creates a new timeout with a callback as our original function.
      // Wait is the number of milliseconds to wait before invoking the callback
      // args is the arguments passed to the callback
      // So if this function is not called again in next number of ms in wait, the original fn is executed.
      timeout = setTimeout(fn, wait, args);
  }
}
export default debounce;

//helper function and classes : END


let audio_blob_array_buffer = null;

// Play file variables
let playing = 0; // 0 not playing, 1 playing
let myInterval = null;
let pollingInterval = 1000; // Value in ms
let operation_alowed = true;

//main functions
window.getPlayFileState  = function(audioFile){
    $.ajax({
        type: 'PUT',
        url: PARAMS.https_audio_Server.concat("/playfile"),
        data: JSON.stringify({ "type": 128, "groups": listOfGroupsToPage, "token": my_id, "file_ID": audioFile.file_name.toString() }),
        processData: false,
        contentType: false,
        complete: function (xhr, textStatus) {
//            console.log("Received response from backend: " + xhr.status)
            if (xhr.status == 200 || playing == 0){
                console.log("File playing has finished.")                
                playStopFile(audioFile)
            }
        }
    });
}

window.playStopFile = function(audioFile){
    let myBtn = document.getElementById("audioFile-button-"+audioFile.id.toString());
    // console.log("Current button ID: " + audioFile.id.toString())
    // Unclick the button
    if ( myBtn.classList.contains("selected") ){
        myBtn.classList.remove("selected");
        //send stop POST request
        playing = 0;
        $.ajax({
            type: 'POST',
            url: PARAMS.https_audio_Server.concat("/playfile"),
            data: JSON.stringify({ "type": 8, "groups": listOfGroupsToPage, "token": my_id, "file_ID": audioFile.file_name.toString() }),
            processData: false,
            contentType: false
        });
        clearInterval(myInterval);
        console.log("File playing stopped");
        operation_alowed = true;
        return;
    }

    // else {
    if (operation_alowed) {
        // Click the button
        myBtn.classList.add("selected");
        operation_alowed = false;
        playing = 1;
        //send start POST request
        $.ajax({
            type: 'POST',
            url: PARAMS.https_audio_Server.concat("/playfile"),
            //Anton the fileID needs to correspond to the file being played
            data: JSON.stringify({ "type": 3, "groups": listOfGroupsToPage, "token": my_id, "file_ID": audioFile.file_name.toString() }),
            processData: false,
            contentType: false,
            success: function (data) {
              console.log("File playing started");
              myInterval = setInterval(getPlayFileState, pollingInterval, audioFile)  // 3rd is the function parameter
//              console.log("---> Status polling started <---")
            },
            error: function (request, status, error) {
              console.log(request, status, error)
              if (error === "Conflict"){
                alert("Conflict: another user is playing at the moment!")
              } else {
                alert(error)
              }
            }
        });
    } else {
      alert("Another announcement is currently played. Please stop it or wait for it to finish.")
    }
}

set_buttons('inactive')
const enablePaging = function (stream) {
  set_buttons('paging_idle');
  var paging_state = 0;
  console.log("Access to microphone ensured");
  const record_button = document.getElementById('record_button');
  const page_button = document.getElementById('page');
  const stop_button = document.getElementById('stop_button');
  const audioContext = new AudioContext();
  const audioSource = audioContext.createMediaStreamSource(stream);
  const audioProcessor = audioContext.createScriptProcessor(PARAMS.bufferLength, PARAMS.channels, PARAMS.channels);
  const resampler = new Resampler(audioContext.sampleRate, PARAMS.sampleRate, PARAMS.channels, PARAMS.bufferLength);
 
  const canvas = document.getElementById("canvas");
  const canvasContext = canvas.getContext("2d");
  const canvasWidth = 80;
  const canvasHeight = 210;

  const processAudioStream = function (audio) {
      let buffer = resampler.resample(audio.inputBuffer.getChannelData(0))
      // let buffer = audio.inputBuffer.getChannelData(0) // No resample, Debugging
      let int16Arr = convertFloat32ToPcm16(buffer, my_id)
      
      $.ajax({
          type: 'PUT',
          url: PARAMS.https_audio_Server.concat("/paging"),
          data: int16Arr.buffer,
          processData: false,
          contentType: false
      });
      
      drawVUMeter(int16Arr);
      drawCanvasLines();
  };

  var paging_state = 0; //0 Not paging, 1 for paging.
  const do_paging = () => {
    if (!playing){ // When there is no file playing
      paging_state = 1;
      operation_alowed = false;  // Prevent file playing once paging has started
      $.ajax({
        type: 'POST',
        url: PARAMS.https_audio_Server.concat("/paging"),
        data: JSON.stringify({ "type": 3, "groups": listOfGroupsToPage, "token": my_id }),
        processData: false,
        contentType: false,
        success: function (data) {
          console.log("Paging started")
          audioSource.connect(audioProcessor)
          audioProcessor.connect(audioContext.destination)
          audioProcessor.onaudioprocess = processAudioStream
          set_buttons('paging_enabled')
        },
        error: function (request, status, error) {
          console.log(request, status, error)
          if (error === "Conflict"){
            alert("Conflict: another user is paging at the moment!")
          } else {
            alert(error)
          }
        }
      });
    }else {
      alert("Wait for any other opepations to finish and try again.")
    }
  };
  
  const stop_paging = debounce(() => {
    if (paging_state == 1) {
      console.log("Paging stopped")
      audioProcessor.onaudioprocess = (audio) => { };
      audioSource.disconnect(audioProcessor);
      audioProcessor.disconnect(audioContext.destination);
      // record_flag = false
      set_buttons('paging_idle')
      paging_state = 0;
      operation_alowed = true;  // Allow file playing once paging is done
      $.ajax({
        type: 'POST',
        url: PARAMS.https_audio_Server.concat("/paging"),
        data: JSON.stringify({ "type": 8, "groups": listOfGroupsToPage, "token": my_id }),
        processData: false,
        contentType: false
      });
      canvasContext.clearRect(0, 0, canvasWidth, canvasHeight);
      drawCanvasLines();
    }
  }, 400);

  // Page button action
  page_button.onmousedown = () => { 
    if (page_button_function === "page"){
      do_paging();
    } 
    else if (page_button_function === "record"){
      startRecording();
    }
    else {
      console.log ('Unknown page_button_function state')
    } 
  };

  page_button.ontouchstart = () => {   
    if (page_button_function === "page"){
      do_paging();
    } 
    else if (page_button_function === "record"){
      startRecording();
    }
    else {
      console.log ('Unknown page_button_function state')
    }   
  };

  page_button.onmouseup = () => {   
    if (page_button_function === "page"){
      stop_paging();
    } 
    else if (page_button_function === "record"){
      stopRecording();
    }
    else {
      console.log ('Unknown page_button_function state')
    }   
  };

  // page_button.onmouseleave = () => {   
  //   if (page_button_function === "page"){
  //     stop_paging();
  //   } 
  //   else if (page_button_function === "record"){
  //     stopRecording();
  //   }
  //   else {
  //     console.log ('Unknown page_button_function state')
  //   }
  // };

  page_button.ontouchend = () => {   
    if (page_button_function === "page"){
      stop_paging();
    } 
    else if (page_button_function === "record"){
      stopRecording();
    }
    else {
      console.log ('Unknown page_button_function state')
    }
  };

  const drawVUMeter = (array) => {
    let average = Math.min(Math.abs(array[1497])*canvasHeight/16000, canvasHeight);
    canvasContext.clearRect(0, 0, canvasWidth, canvasHeight);
    canvasContext.fillStyle = '#BadA55';
    canvasContext.fillRect(0, canvasHeight - average, canvasWidth, canvasHeight);
  };

};

const drawCanvasLines = () => {

  const canvas = document.getElementById("canvas");
  const canvasContext = canvas.getContext("2d");
  const canvasWidth = 80;
  const canvasHeight = 210;

  let y = 15;
  while (y < canvasHeight){
    canvasContext.beginPath();
    canvasContext.moveTo(0, y);
    canvasContext.lineTo(canvasWidth, y);
    canvasContext.stroke();
    y += 15;
  }
}

const make_connection = () =>{
  PARAMS.sampleRate = document.getElementById("sample_rate_val_id").textContent;  
  const urlParams = new URLSearchParams(window.location.search);
  my_id = urlParams.get('token');
  // Create an audio stream, then call enablePaging with it
  navigator.mediaDevices.getUserMedia({ audio: true, video: false }).then(enablePaging);
  set_buttons('paging_idle');
};

const release_connection = () =>{
  set_buttons('inactive')
};

let listOfGroupsToPage = [];
const selectOrDeselectGroup = (group) => {
  const groupBtnElem = document.getElementById("group-button-"+group.id.toString());
  const index = listOfGroupsToPage.indexOf(group.id);
  if (index > -1) {
    // deselect
    listOfGroupsToPage.splice(index, 1);
    groupBtnElem.setAttribute("style", "background-color: #3494b4;");
  } else {
    //select
    listOfGroupsToPage.push(group.id);
    groupBtnElem.setAttribute("style", "background-color: #0037ff;");
  }
  const selectAllCheckbox = document.getElementById("selectAllCheckbox");
  if (allGroupsSelected()){
    selectAllCheckbox.checked = true;
  } else {
    selectAllCheckbox.checked = false;
  }
};

const selectAllGroupsOnClick = () => {
  const selectAllCheckbox = document.getElementById("selectAllCheckbox");
  if (selectAllCheckbox.checked){ //select all
    listOfGroupsToPage = [];
    for (let i = 0; i < sortedGroupsList.length; i++){
      listOfGroupsToPage.push(sortedGroupsList[i].id);
      const groupBtnElem = document.getElementById("group-button-"+sortedGroupsList[i].id.toString());
      groupBtnElem.setAttribute("style", "background-color: #0037ff;");
    }
  } else { //select none
    listOfGroupsToPage = [];
    for (let i = 0; i < sortedGroupsList.length; i++){
      const groupBtnElem = document.getElementById("group-button-"+sortedGroupsList[i].id.toString());
      groupBtnElem.setAttribute("style", "background-color: #3494b4;");
    }
  }
}

function allGroupsSelected() {
  let allChecked = true;
  for (let i = 0; i < sortedGroupsList.length; i++){
    const index = listOfGroupsToPage.indexOf(sortedGroupsList[i].id);
    if (index === -1) {
      allChecked = false;
      break;
    }
  }
  return allChecked;
}

window.onload = function () {
  set_buttons('inactive')
  make_connection();
  drawCanvasLines();
  
  //groups
  loadGroups(selectOrDeselectGroup);

  const selectAllCheckbox = document.getElementById('selectAllCheckbox');
  selectAllCheckbox.onclick = function () {
    selectAllGroupsOnClick();
  }
  if (window.screen.width < 500){
    selectAllCheckbox.setAttribute("style", "width: 40px; height: 40px;");
  } else {
    selectAllCheckbox.setAttribute("style", "width: 30px; height: 30px;");
  }
  const selectAllSpan = document.getElementById("selectAllSpan");
  if (window.screen.width < 500){
    selectAllSpan.setAttribute("style", "font-size: 30px;");
  } else {
    selectAllSpan.setAttribute("style", "font-size: 24px;");
  }  

  // Recording functionality
  const enableRecordingCheckbox = document.getElementById('enableRecordingCheckbox');
  enableRecordingCheckbox.onclick = function () {
      enableRecordingOnClick();
  }
  if (window.screen.width < 500){
      enableRecordingCheckbox.setAttribute("style", "width: 20px; height: 20px;");
  } else {
      enableRecordingCheckbox.setAttribute("style", "width: 15px; height: 15px;");
  }

  // Prerecorded messages
  loadAudioFiles(playStopFile);  // Each of the audio buttons would be calling playStopFile function
}

 

// START OF AUDIO RECORDING
// Record message from browser
const recording = document.getElementById("recording");
const logElement = document.getElementById("log_recording");
const sendButton = document.getElementById("sendButton");

navigator.getUserMedia = navigator.getUserMedia || 
                         navigator.webkitGetUserMedia || 
                         navigator.mozGetUserMedia || 
                         navigator.msGetUserMedia;
 
let audioStreamNew = null; 
const leftChannel = [];
let recordingLength = 0;
let recorder;
let volume;
let audioInput;
let recordingResampler;
let smp_rate = -1;
let recording_active; // 0 idle, 1 paging
let mediaRecorderData = [];  // used for the preview
let recordedBlob = null;

const enableRecordingOnClick = () => {
  const enableRecordingCheckbox = document.getElementById("enableRecordingCheckbox");
  if (enableRecordingCheckbox.checked){ //enable recording
    // Change the function of the PAGE button
    set_buttons("recording_idle")
    // Display the record and send buttons
    document.getElementById("recording").style.display="block"
    document.getElementById("log_recording").style.display="block"
    document.getElementById("sendButton").style.display="block"
  } else { //set back to paging   
    // Change the function of the PAGE button 
    set_buttons("paging_idle")
    // Hide the record and send buttons
    document.getElementById("recording").style.display="none"
    document.getElementById("log_recording").style.display="none"
    document.getElementById("sendButton").style.display="none"

    leftChannel.length = 0;
    recordingLength = 0;
    logElement.innerHTML = null;
    recording.src = ``;
  }
  // console.log("Changed page-button function: " + page_button_function)
}


function log_recording(msg) {
  logElement.innerHTML += msg + "\n";
}

async function startRecording () {

  let stream = null;
  const constraints = {audio: true, video: false};
  
  if (navigator.mediaDevices.getSupportedConstraints() != null){
      // Prevent pressing twice on the start button
      if (!recording_active) {
          recording_active = 1;

          try {
              stream = await navigator.mediaDevices.getUserMedia(constraints);
              // Make sure the buffer is empty when a new recording is started          
              set_buttons('recording_start');
              leftChannel.length = 0;
              recordingLength = 0;

              audioStreamNew = stream;

              log_recording("Recording started...")
              // creates the an instance of audioContext
              const context = window.AudioContext || window.webkitAudioContext;
              const audioContext = new context();
              
              // retrieve the current sample rate of microphone the browser is using
              const sampleRate = audioContext.sampleRate;
              smp_rate = audioContext.sampleRate
              
              // creates a gain node
              volume = audioContext.createGain();
              
              // creates an audio node from the microphone incoming stream
              audioInput = audioContext.createMediaStreamSource(audioStreamNew);
              
              // connect the stream to the gain node
              audioInput.connect(volume);

              /* Controls how frequently the audioprocess event is dispatched and 
              how many sample-frames need to be processed each call.
              Lower values for buffer size will result in a lower (better) latency.
              Higher values will be necessary to avoid audio breakup and glitches */
              const bufferSize = 2 * 1024;
              recorder = (audioContext.createScriptProcessor || 
                          audioContext.createJavaScriptNode).call(audioContext, 
                                                                  bufferSize, 
                                                                  1, 
                                                                  1); 
              // we connect the recorder
              volume.connect(recorder);
              
              // start recording
              recorder.connect(audioContext.destination);    

              recorder.onaudioprocess = function(event){
                  const samples = event.inputBuffer.getChannelData(0);
                  // we clone the samples
                  leftChannel.push(new Float32Array(samples));
                  // increase the recorded length
                  recordingLength += bufferSize;
              };
              
              //TODO: Create a blob here to hold the data for the preview
              mediaRecorderData = await recordPreview(stream);
              recordedBlob = new Blob(mediaRecorderData, { type: "audio" });
              // console.log("Recording data:")
              // console.log(recordedBlob)  
              recording.src = URL.createObjectURL(recordedBlob);        

          } catch (err) {
            // Handle the error here
            console.log(`Couldn't capture the media stream. Error: ${err.toString()}`)
          }  
      } else {
          console.log("Message is currently being recorded")
      }       
  }  
  else {
        alert('getUserMedia not supported in this browser');
  }  
}

async function recordPreview (stream) {
  let recorder = new MediaRecorder(stream);
  let data = [];

  recorder.ondataavailable = event => data.push(event.data);
  recorder.start();
  // log_recording("Recorder state " + recorder.state);

  let stopped = new Promise((resolve, reject) => {
    recorder.onstop = resolve;
    recorder.onerror = event => reject(event.name);
  });

  await Promise.all([
    stopped
  ]);
  return data;
}


const stopRecording = debounce(() => {
  recorder.disconnect();  
  volume.disconnect();
  audioInput.disconnect();
  if (audioStreamNew !== undefined) {
    audioStreamNew.getTracks().forEach(track => {
          track.stop();
      });
  }
  // console.log("Recording stopped")
  log_recording("Recording stopped")
  set_buttons('recording_idle');
  recording_active = 0;
}, 400)
                


function mergeBuffers(channelBuffer, recordingLength)
{
   let result = new Float32Array(recordingLength);
   let offset = 0;
 
   for (let i = 0; i < channelBuffer.length; i++)
   {
       result.set(channelBuffer[i], offset);
       offset += channelBuffer[i].length;
   }
 
   return Array.prototype.slice.call(result);
}


function attachID(inputBuffer,recordingLength ,my_token) {
  const PCM32fSamples = mergeBuffers(inputBuffer, recordingLength);
  // const PCM16iSamples = [];  // This buffer holds the PCM 16 audio data
  recordingResampler = new Resampler(smp_rate, PARAMS.sampleRate, PARAMS.channels, recordingLength);
  let buffer = recordingResampler.resample(PCM32fSamples)


  let l = buffer.length + my_token.length;  // So we can add the token in the beginning
  const PCM16iSamples = new Int16Array(l);  // This buffer holds the PCM 16 audio data

  // console.log("Current token: " + my_token)

  for (let idx = 0; idx < my_token.length; idx++) {
    // PCM16iSamples.push(my_token.charCodeAt(idx));
    PCM16iSamples[idx] = my_token.charCodeAt(idx)
  }

  for (let i = my_token.length; i < l; i++)
  {
    let val = Math.floor(32767 * buffer[i-my_token.length]);
    val = Math.min(32767, val);
    val = Math.max(-32768, val);
  
    // PCM16iSamples.push(val);
    PCM16iSamples[i] = val
  }

  // console.log(PCM16iSamples);
  return PCM16iSamples;
}
                            
sendButton.addEventListener("click", function() {
  sendAudioFile(attachID(leftChannel, recordingLength, my_id));
},false);


window.sendAudioFile  = function(audioFile){

  // console.log("Type of data to send: " + typeof(audioFile))
  // console.log(audioFile)
  log_recording("Paging recording, please wait...")

  $.ajax({
    type: 'POST',
    url: PARAMS.https_audio_Server.concat("/recording"),
    data: JSON.stringify({ "type": 3, "groups": listOfGroupsToPage, "token": my_id }),
    processData: false,
    contentType: false,
    success: function (data) {
      console.log("Authorised. Sending data...")
      $.ajax({
        type: 'PUT',
        url: PARAMS.https_audio_Server.concat("/recording"),
        data: audioFile,
        processData: false,
        contentType: false,
        success: function(data) {
          log_recording("Recording paged to group: " + listOfGroupsToPage)
          console.log("Data sending completed.")
          $.ajax({
            type: 'POST',
            url: PARAMS.https_audio_Server.concat("/recording"),
            data: JSON.stringify({ "type": 8, "groups": listOfGroupsToPage, "token": my_id }),
            processData: false,
            contentType: false
          });
        },
        error: function (request, status, error) {
          console.log(request, status, error)
          $.ajax({
            type: 'POST',
            url: PARAMS.https_audio_Server.concat("/recording"),
            data: JSON.stringify({ "type": 8, "groups": listOfGroupsToPage, "token": my_id }),
            processData: false,
            contentType: false
          });
        }
      });
    },
    error: function (request, status, error) {
      console.log(request, status, error)
      if (error === "Conflict"){
        alert("Conflict: another user is paging at the moment!")
      } else {
        alert(error)
      }
    }
  });
}